#include <amxmodx>
#include <engine>
#include <fakemeta>

#pragma semicolon				1

#define PLUGIN_NAME				"Grab"
#define PLUGIN_VERS				"0.2"
#define PLUGIN_AUTH				"81x08"

#define MAX_PLAYERS				32

#define IsEntity(%0)			is_valid_ent(%0)
#define IsPlayer(%0)			(1 <= %0 <= 32)

#define GRAB_FLAG_ACCESS		ADMIN_KICK
#define GRAB_FLAG_IMMUNITY		ADMIN_RCON

#define GRAB_OPTIMIZED

#if defined GRAB_OPTIMIZED
	#define GRAB_THINK_DELAY	0.01
#else
	#include <hamsandwich>
#endif

#define IN_PUSH					(1<<0)
#define IN_PULL					(1<<1)

enum _: ENUM_DATA_PLYAER_GRAB	{
	PL_GRAB_READY = -1,
	PL_GRAB_NULL,

	Float: PL_GRAB_LEN,
	
	PL_GRAB_FLAGS,
	PL_GRAB_OWNER,
	PL_GRAB_TARGET
};

#if defined GRAB_OPTIMIZED
	new gp_iThinkGrab[MAX_PLAYERS + 1 char];
#endif

new gp_iAccess[MAX_PLAYERS + 1 char],
	gp_iImmunity[MAX_PLAYERS + 1 char],
	gp_iGrab[MAX_PLAYERS + 1 char][ENUM_DATA_PLYAER_GRAB];

/*================================================================================
 [PLUGIN]
=================================================================================*/
public plugin_init()	{
	/* [PLUGIN] */
	register_plugin(PLUGIN_NAME, PLUGIN_VERS, PLUGIN_AUTH);

	/* [THINK | HAMSANDWICH] */
	#if defined GRAB_OPTIMIZED
		register_think("think_grab", "ThinkHook_Grab");
	#else
		RegisterHam(Ham_Player_PreThink, "player", "HamHook_Player_PreThink", true);
	#endif

	/* [CLCMD] */
	register_clcmd("drop", "ClCmd_Throw");

	register_clcmd("+grab", "ClCmd_GrabOn");
	register_clcmd("-grab", "ClCmd_GrabOff");

	register_clcmd("+pull", "ClCmd_PullOn");
	register_clcmd("-pull", "ClCmd_PullOff");

	register_clcmd("+push", "ClCmd_PushOn");
	register_clcmd("-push", "ClCmd_PushOff");
}

/*================================================================================
 [CLIENT]
=================================================================================*/
public client_putinserver(pId)	{
	new iFlags = get_user_flags(pId);

	if(iFlags & GRAB_FLAG_ACCESS) gp_iAccess[pId] = true;
	if(iFlags & GRAB_FLAG_IMMUNITY) gp_iImmunity[pId] = true;
	
	#if defined GRAB_OPTIMIZED
		if((gp_iThinkGrab[pId] = create_entity("info_target")))
			entity_set_string(gp_iThinkGrab[pId], EV_SZ_classname, "think_grab");
	#endif
}

public client_disconnect(pId)	{
	gp_iImmunity[pId] = false;

	if(gp_iAccess[pId])	{
		gp_iAccess[pId] = false;
		
		#if defined GRAB_OPTIMIZED
			if(gp_iThinkGrab[pId])	{
				entity_set_int(gp_iThinkGrab[pId], EV_INT_flags, entity_get_int(gp_iThinkGrab[pId], EV_INT_flags) | FL_KILLME);
				
				gp_iThinkGrab[pId] = 0;
			}
		#endif
	}
}

/*================================================================================
 [THINK]
=================================================================================*/
public ThinkHook_Grab(const eId)	{
	jb_think_grab(entity_get_int(eId, EV_INT_iuser1));
	entity_set_float(eId, EV_FL_nextthink, get_gametime() + GRAB_THINK_DELAY);
}

/*================================================================================
 [HAMSANDWICH]
=================================================================================*/
public HamHook_Player_PreThink(const pId)	{
	if(gp_iGrab[pId][PL_GRAB_TARGET] == PL_GRAB_READY)	{
		new iTarget = UTIL_GetEntityAiming(pId);
		if(iTarget) jb_set_grab(pId, iTarget);
	} else if(gp_iGrab[pId][PL_GRAB_TARGET]) jb_think_grab(pId);
}

/*================================================================================
 [CLCMD]
=================================================================================*/
public ClCmd_Throw(const pId)	{
	new iTarget = gp_iGrab[pId][PL_GRAB_TARGET];
	if(iTarget)	{
		entity_set_vector(iTarget, EV_VEC_velocity, UTIL_VelocityByAim(pId, 1500.0));
		jb_unset_grab(pId);
		return PLUGIN_HANDLED;
	}
	
	return PLUGIN_CONTINUE;
}

public ClCmd_GrabOn(const pId)	{
	if(!(gp_iAccess[pId]))
		return PLUGIN_HANDLED;
	
	#if defined GRAB_OPTIMIZED
		jb_set_grab(pId, UTIL_GetEntityAiming(pId));
		
		entity_set_int(gp_iThinkGrab[pId], EV_INT_iuser1, pId);
		entity_set_float(gp_iThinkGrab[pId], EV_FL_nextthink, get_gametime() + GRAB_THINK_DELAY);
	#else
		gp_iGrab[pId][PL_GRAB_TARGET] = PL_GRAB_READY;
	#endif

	return PLUGIN_HANDLED;
}

public ClCmd_GrabOff(const pId)	{
	jb_unset_grab(pId);

	return PLUGIN_HANDLED;
}

public ClCmd_PullOn(const pId)	{
	if(gp_iGrab[pId][PL_GRAB_TARGET])
		gp_iGrab[pId][PL_GRAB_FLAGS] ^= IN_PULL;

	return PLUGIN_HANDLED;
}

public ClCmd_PullOff(const pId)	{
	if(gp_iGrab[pId][PL_GRAB_FLAGS] & IN_PULL)
		gp_iGrab[pId][PL_GRAB_FLAGS] &= ~IN_PULL;

	return PLUGIN_HANDLED;
}

public ClCmd_PushOn(const pId)	{
	if(gp_iGrab[pId][PL_GRAB_TARGET])
		gp_iGrab[pId][PL_GRAB_FLAGS] ^= IN_PUSH;
	
	return PLUGIN_HANDLED;
}

public ClCmd_PushOff(const pId)	{
	if(gp_iGrab[pId][PL_GRAB_FLAGS] & IN_PUSH)
		gp_iGrab[pId][PL_GRAB_FLAGS] &= ~IN_PUSH;

	return PLUGIN_HANDLED;
}

/*================================================================================
 [STOCK]
=================================================================================*/
static stock jb_set_grab(const pId, const iTarget)	{
	if(gp_iGrab[pId][PL_GRAB_OWNER])
		return PLUGIN_HANDLED;

	gp_iGrab[pId][PL_GRAB_TARGET] = iTarget;

	if(IsPlayer(iTarget))	{
		if(gp_iGrab[iTarget][PL_GRAB_OWNER] || gp_iGrab[iTarget][PL_GRAB_TARGET])	{
			gp_iGrab[pId][PL_GRAB_TARGET] = PL_GRAB_NULL;
			return PLUGIN_HANDLED;
		}

		gp_iGrab[iTarget][PL_GRAB_OWNER] = pId;
	} else {
		if(entity_get_int(iTarget, EV_INT_iuser1))
			return PLUGIN_HANDLED;

		entity_set_int(iTarget, EV_INT_iuser1, pId);
	}

	new Float: fOrigin[3], Float: fTargetOrigin[3];
	entity_get_vector(pId, EV_VEC_origin, fOrigin);
	entity_get_vector(iTarget, EV_VEC_origin, fTargetOrigin);

	gp_iGrab[pId][PL_GRAB_LEN] = _:get_distance_f(fOrigin, fTargetOrigin);
	if(gp_iGrab[pId][PL_GRAB_LEN] < 90.0) gp_iGrab[pId][PL_GRAB_LEN] = _:90.0;

	new Float: fColor[3];
	fColor[0] = random_float(0.0, 255.0);
	fColor[1] = random_float(0.0, 255.0);
	fColor[2] = random_float(0.0, 255.0);

	UTIL_SetEntityRendering(pId, kRenderFxGlowShell, fColor, kRenderNormal, 35.0);
	UTIL_SetEntityRendering(iTarget, kRenderFxGlowShell, fColor, kRenderNormal, 35.0);
	
	return PLUGIN_HANDLED;
}

static stock jb_unset_grab(const pId)	{
	if(gp_iGrab[pId][PL_GRAB_TARGET] == PL_GRAB_READY) gp_iGrab[pId][PL_GRAB_TARGET] = PL_GRAB_NULL;
	else {
		UTIL_SetEntityRendering(pId, kRenderFxNone, Float: {0.0, 0.0, 0.0}, kRenderNormal, 0.0);

		if(IsPlayer(gp_iGrab[pId][PL_GRAB_TARGET]))	{
			gp_iGrab[gp_iGrab[pId][PL_GRAB_TARGET]][PL_GRAB_OWNER] = PL_GRAB_NULL;
			UTIL_SetEntityRendering(gp_iGrab[pId][PL_GRAB_TARGET], kRenderFxNone, Float: {0.0, 0.0, 0.0}, kRenderNormal, 0.0);
		} else if(IsEntity(gp_iGrab[pId][PL_GRAB_TARGET]))	{
			entity_set_int(gp_iGrab[pId][PL_GRAB_TARGET], EV_INT_iuser1, PL_GRAB_NULL);
			UTIL_SetEntityRendering(gp_iGrab[pId][PL_GRAB_TARGET], kRenderFxNone, Float: {0.0, 0.0, 0.0}, kRenderNormal, 0.0);
		}

		gp_iGrab[pId][PL_GRAB_TARGET] = PL_GRAB_NULL;
		
		#if defined GRAB_OPTIMIZED
			entity_set_float(gp_iThinkGrab[pId], EV_FL_nextthink, 0.0);
		#endif
	}
}

static stock jb_think_grab(const pId)	{
	new iTarget = gp_iGrab[pId][PL_GRAB_TARGET];
	if(!(IsEntity(iTarget)))	{
		jb_unset_grab(pId);
		return;
	}

	new Float: fViewOfs[3], Float: fVelocity[3], Float: fOrigin[3], Float: fReturn[3];
	
	fViewOfs = UTIL_GetViewOfs(pId);
	fVelocity = UTIL_VelocityByAim(pId, gp_iGrab[pId][PL_GRAB_LEN]);

	entity_get_vector(iTarget, EV_VEC_origin, fOrigin);

	fReturn[0] = ((fViewOfs[0] + fVelocity[0]) - fOrigin[0]) * 8;
	fReturn[1] = ((fViewOfs[1] + fVelocity[1]) - fOrigin[1]) * 8;
	fReturn[2] = ((fViewOfs[2] + fVelocity[2]) - fOrigin[2]) * 8;

	entity_set_vector(iTarget, EV_VEC_velocity, fReturn);
	
	static iButton; iButton = entity_get_int(pId, EV_INT_button);

	if(iButton & IN_ATTACK || gp_iGrab[pId][PL_GRAB_FLAGS] & IN_PULL)	{
		new Float: iLen = gp_iGrab[pId][PL_GRAB_LEN];
		if(iLen > 90.0)	{
			iLen -= 5.0;
			iLen < 90.0 ? (iLen = 90.0) : (gp_iGrab[pId][PL_GRAB_LEN] = _:iLen);
		}
	}
	
	if(iButton & IN_ATTACK2 || gp_iGrab[pId][PL_GRAB_FLAGS] & IN_PUSH)
		gp_iGrab[pId][PL_GRAB_LEN] += 5.0;
}

/*================================================================================
 [UTIL]
=================================================================================*/
static stock UTIL_GetEntityAiming(const pId)	{
	new Float: fAngles[3]; fAngles = UTIL_VelocityByAim(pId, 9999.0);
	new Float: fViewOfs[3]; fViewOfs = UTIL_GetViewOfs(pId);

	fAngles[0] += fViewOfs[0];
	fAngles[1] += fViewOfs[1];
	fAngles[2] += fViewOfs[2];

	engfunc(EngFunc_TraceLine, fViewOfs, fAngles, DONT_IGNORE_MONSTERS, pId, 0);

	new iTarget = get_tr2(0, TR_pHit);

	if(!(IsPlayer(iTarget)))	{
		iTarget = 0;
		
		new Float: fEndPos[3];
		get_tr2(0, TR_vecEndPos, fEndPos);

		new eId = find_ent_in_sphere(-1, fEndPos, 12.0), iMovetype;
		while(!(iTarget) && eId > 0)	{
			iMovetype = entity_get_int(eId, EV_INT_movetype);
			if((iMovetype == MOVETYPE_WALK || iMovetype == MOVETYPE_STEP || iMovetype == MOVETYPE_TOSS) && eId != pId) iTarget = eId;
			else eId = find_ent_in_sphere(eId, fEndPos, 12.0);
		}
	} else if(gp_iImmunity[iTarget]) iTarget = PL_GRAB_NULL;

	return iTarget;
}

static stock Float: UTIL_GetViewOfs(const pId)	{
	new Float: fOrigin[3], Float: fViewOfs[3];
	entity_get_vector(pId, EV_VEC_origin, fOrigin);
	entity_get_vector(pId, EV_VEC_view_ofs, fViewOfs);

	fOrigin[0] += fViewOfs[0];
	fOrigin[1] += fViewOfs[1];
	fOrigin[2] += fViewOfs[2];

	return fOrigin;
}

static stock Float: UTIL_VelocityByAim(const pId, const Float: fSpeed)	{
	new Float: fAngles[3];
	entity_get_vector(pId, EV_VEC_v_angle, fAngles);
	engfunc(EngFunc_MakeVectors, fAngles);
	global_get(glb_v_forward, fAngles);

	fAngles[0] *= fSpeed;
	fAngles[1] *= fSpeed;
	fAngles[2] *= fSpeed;
	
	return fAngles;
}

static stock UTIL_SetEntityRendering(const pId, const iRenderFx, const Float: fColor[3], const iRenderMode, const Float: fRenderAmt)	{
	entity_set_int(pId, EV_INT_renderfx, iRenderFx);

	entity_set_vector(pId, EV_VEC_rendercolor, fColor);
	
	entity_set_int(pId, EV_INT_rendermode, iRenderMode);
	
	entity_set_float(pId, EV_FL_renderamt, fRenderAmt);
}